//floordata.c

#ifdef __cplusplus
extern "C"
{
#endif

#include <stdio.h>
#include <string.h>
#include <crtdbg.h>
#include "MyMemory.h"
#include "trmdefs.h"
#include "trldefs.h"
#include "floordata.h"
#include "linklist.h"

#define ASSERT(e) _ASSERT(e)

floordatalist_t *floordatahead = NULL;
floordatalist_t *floordatatail = NULL;
short g_fdindex = 0;

trm_floordata_t *FloorDataAdd(void *data, short size)
{
    trm_floordata_t *fd;
    floordatalist_t *f;
    ASSERT(size <= MAX_FLOORDATA_SIZE);
    f = (floordatalist_t *)NewListItem(NULL, sizeof(trm_floordata_t), (linklist_t **)&floordatahead, (linklist_t **)&floordatatail);
    fd = f->floordata;
    memcpy(fd->floordata, data, size);
    fd->size = size;
    fd->index = g_fdindex;
    g_fdindex += size/2;
    return fd;
}

void FreeFloorDataList(void)
{
    linklist_t *l, *l2;
    l = (linklist_t *)floordatahead;
    while (l)
    {
        l2 = l->next;
        FreeListItem(l, sizeof(trm_floordata_t));
        l = l2;
    }
    floordatahead = NULL;
    floordatatail = NULL;
    g_fdindex = 0;
}

void BuildFloorData(trm_t *ptrm)
{
    unsigned char buf[MAX_FLOORDATA_SIZE];
    trm_sector_t *sector;
    trm_roomlist_t *r;
    trm_room_t *room;
    trm_floordata_t *fd;
    floordatalist_t *f;
    long fdcount = 0, finalsize;
    int c, x;
    short size;
    buf[0] = 0;
    buf[1] = 0;
    FloorDataAdd(buf, 2);
    //fixme: allrooms
    r = ptrm->roomlisthead;
    for (c = 0; c < ptrm->numrooms; c++)
    {
        room = r->room;
        for (x = 0; x < room->numxsectors*room->numysectors; x++)
        {
            size = 0;
            sector = &room->sectors[x];
            if ((sector->lrfloor != 0) || (sector->udfloor != 0))
            {
                buf[size++] = 2;
                buf[size++] = 0x80;
                buf[size++] = sector->lrfloor;
                buf[size++] = sector->udfloor;
            }
            if ((sector->lrceiling != 0) || (sector->udceiling != 0))
            {
                if (size)
                {
                    buf[1] = 0;
                }
                buf[size++] = 3;
                buf[size++] = 0x80;
                buf[size++] = sector->lrceiling;
                buf[size++] = sector->udceiling;
            }
            if (sector->doorroom != -1)
            {
                ASSERT(!size);
                buf[size++] = 1;
                buf[size++] = 0x80;
                *(short *)(buf+size) = sector->doorroom;
                size += 2;
            }
            if (size)
            {
                fd = FloorDataAdd(buf, size);
                sector->fdindex = fd->index;
            }
        }
        r = r->next;
    }
    f = floordatahead;
    finalsize = 0;
    while (f)
    {
        finalsize += f->floordata->size;
        f = f->next;
    }
    if (!finalsize)
    {
        ptrm->numfloordata = 0;
        FreeFloorDataList();
        return;
    }
    ptrm->numfloordata = finalsize/2;
    ptrm->floordata = (short *)GetMemBlock(finalsize);
    f = floordatahead;
    finalsize = 0;
    while (f)
    {
        memcpy(((unsigned char *)ptrm->floordata)+finalsize, f->floordata->floordata, f->floordata->size);
        finalsize += f->floordata->size;
        f = f->next;
    }
    FreeFloorDataList();
}

#ifdef __cplusplus
}
#endif
