//room.c

#ifdef __cplusplus
extern "C"
{
#endif

#include <stdio.h>
#include <string.h>
#include <crtdbg.h>
#include "types.h"
#include "MyMemory.h"
#include "trdefs.h"
#include "trmdefs.h"
#include "room.h"
#include "linklist.h"

#define ASSERT(e) _ASSERT(e)

trm_room_t *NewRoom(trm_t *trm, short numxsectors, short numysectors)
{
    trm_room_t *room;
    trm_sector_t *sectors;
    trm_roomlist_t *r;
    int x, y, i;
    r = (trm_roomlist_t *)NewListItem(NULL, sizeof(trm_room_t), (linklist_t **)&trm->roomlisthead, (linklist_t **)&trm->roomlisttail);
    room = r->room;
    memset(room, 0, sizeof(trm_room_t));
    room->numxsectors = numxsectors;
    room->numysectors = numysectors;
    sectors = room->sectors = (trm_sector_t *)GetMemBlock((numxsectors*numysectors)*sizeof(trm_sector_t));
    for (y = 0; y < numysectors; y++)
    {
        sectors[y].fdindex = 0;
        sectors[y].camera = -1;
        sectors[y].floorattrib = -1;
        sectors[y].doorroom = -1;
        sectors[y].floor = SECTOR_WALL;
        sectors[y].ceilingattrib = -1;
        sectors[y].ceiling = SECTOR_WALL;
        sectors[y].lrceiling = 0;
        sectors[y].udceiling = 0;
        sectors[y].lrfloor = 0;
        sectors[y].udfloor = 0;
    }
    for (x = 1, i = numysectors; x < numxsectors-1; x++)
    {
        for (y = 0; y < numysectors; y++, i++)
        {
            sectors[i].fdindex = 0;
            sectors[i].camera = -1;
            sectors[i].floorattrib = -1;
            sectors[i].ceilingattrib = -1;
            sectors[i].lrceiling = 0;
            sectors[i].udceiling = 0;
            sectors[i].lrfloor = 0;
            sectors[i].udfloor = 0;
            sectors[i].doorroom = -1;
            if ((y == 0) || (y == numysectors-1))
            {
                sectors[i].floor = SECTOR_WALL;
                sectors[i].ceiling = SECTOR_WALL;
            }
            else
            {
                //normal sectors
                if ((x%2 == 0) || (y%2 == 0))
                {
                    sectors[i].floor = 0;
                    sectors[i].ceiling = -8;
                }
                else
                {
                    sectors[i].floor = 0;
                    sectors[i].ceiling = -8;
                    //sectors[i].lrfloor = -1;
                    //sectors[i].udfloor = -1;
                }
            }
        }
    }
    for (y = 0; y < numysectors; y++, i++)
    {
        sectors[i].fdindex = 0;
        sectors[i].camera = -1;
        sectors[i].floorattrib = -1;
        sectors[i].floor = SECTOR_WALL;
        sectors[i].ceilingattrib = -1;
        sectors[i].ceiling = SECTOR_WALL;
        sectors[i].lrceiling = 0;
        sectors[i].udceiling = 0;
        sectors[i].lrfloor = 0;
        sectors[i].udfloor = 0;
        sectors[i].doorroom = -1;
    }
    trm->numrooms++;
    trm->roomcurrent = r;
    return room;
}

void ChangeRoomSize(trm_room_t *room, int newx, int newy)
{
    trm_sector_t *sectors;
    int x, y, xsize, ysize;
    if (newx == -1)
    {
        xsize = room->numxsectors;
    }
    else
    {
        xsize = newx;
    }
    if (newy == -1)
    {
        ysize = room->numysectors;
    }
    else
    {
        ysize = newy;
    }
    sectors = (trm_sector_t *)GetMemBlock((xsize*ysize)*sizeof(trm_sector_t));
}

#ifdef __cplusplus
}
#endif
