//trl.c

#ifdef __cplusplus
extern "C"
{
#endif

#include <stdio.h>
#include <string.h>
#include <crtdbg.h>
#include "types.h"
#include "trmdefs.h"
#include "trldefs.h"
#include "trl.h"
#include "floordata.h"

#define ASSERT(e) _ASSERT(e)

long BuildTrlFile(char *lpszFileName, trm_t *ptrm)
{
    trl_header_t header;
    sector_t dsector;
    FILE *f;
    trm_room_t *room;
    trm_roomlist_t *r;
    trm_sector_t *ssector;
    trm_itemlist_t *i;
    int c, x;
    short s;
    //backup old lpszFileName?
    BuildFloorData(ptrm);
    memset(&header, 0, sizeof(trl_header_t));
    memcpy(header.magic, TRL_MAGIC, 4);
    header.trlversion = TRL_VERSION;
    header.gameversion = TR2_VERSION;
    memcpy(header.baselevel, ptrm->header.baselevel, strlen(ptrm->header.baselevel));
    memcpy(header.levelname, ptrm->header.levelname, strlen(ptrm->header.levelname));
    f = fopen(lpszFileName, "w+b");
    fseek(f, sizeof(trl_header_t), SEEK_SET);
    header.dir.o_textiles = ftell(f);
    fwrite(&ptrm->numtextiles, 4, 1, f);
    ASSERT(!ptrm->numtextiles);
    if (ptrm->numtextiles)
    {
        fwrite(ptrm->textiles8, sizeof(textile8_t), ptrm->numtextiles, f);
        fwrite(ptrm->textiles16, sizeof(textile16_t), ptrm->numtextiles, f);
    }
    header.dir.s_textiles = ftell(f)-header.dir.o_textiles;
    header.dir.o_rooms = ftell(f);
    fwrite(&ptrm->numrooms, 2, 1, f);
    r = ptrm->roomlisthead;
    for (c = 0; c < ptrm->numrooms; c++)
    {
        room = r->room;
        fwrite(&room->roominfo, sizeof(roominfo_t), 1, f);
        fwrite(&room->nummeshwords, 4, 1, f);
        ASSERT(room->nummeshwords);
        //fixme
        //fwrite(room->meshwords, room->nummeshwords, 2, f);
        fwrite(&room->mesh.numvertices, 2, 1, f);
        if (room->mesh.numvertices > 0)
        {    
            fwrite(room->mesh.vertices, sizeof(rmvertex2_t), room->mesh.numvertices, f);
        }
        fwrite(&room->mesh.numpoly4s, 2, 1, f);
        if (room->mesh.poly4s > 0)
        {    
            fwrite(room->mesh.poly4s, sizeof(poly4_t), room->mesh.numpoly4s, f);
        }
        fwrite(&room->mesh.numpoly3s, 2, 1, f);
        if (room->mesh.poly3s > 0)
        {    
            fwrite(room->mesh.poly3s, sizeof(poly3_t), room->mesh.numpoly3s, f);
        }
        fwrite(&room->mesh.un1, 2, 1, f);

        fwrite(&room->numdoors, 2, 1, f);
        if (room->numdoors)
        {    
            fwrite(room->doors, sizeof(roomdoor_t), room->numdoors, f);
        }
        fwrite(&room->numysectors, 2, 1, f);
        fwrite(&room->numxsectors, 2, 1, f);
        for (x = 0; x < room->numxsectors*room->numysectors; x++)
        {
            ssector = &room->sectors[x];
            dsector.fdindex = ssector->fdindex;
            dsector.camera = ssector->camera;
            dsector.floorattrib = ssector->floorattrib;
            dsector.floor = ssector->floor;
            dsector.ceilingattrib = ssector->ceilingattrib;
            dsector.ceiling = ssector->ceiling;
            fwrite(&dsector, sizeof(sector_t), 1, f);
        }
        fwrite(&room->un1, 1, 1, f);
        fwrite(&room->lightlevel1, 1, 1, f);
        fwrite(&room->un4, 1, 1, f);
        fwrite(&room->lightlevel2, 1, 1, f);
        s = 0;
        fwrite(&s, 2, 1, f);
        fwrite(&s, 2, 1, f);
        fwrite(&s, 2, 1, f);
    
        room->un2 = -1;
        fwrite(&room->un2, 2, 1, f);
        fwrite(&room->water, 1, 1, f);
        fwrite(&room->un3, 1, 1, f);
        r = r->next;
    }
    header.dir.s_rooms = ftell(f)-header.dir.o_rooms;
    header.dir.o_floordata = ftell(f);
    fwrite(&ptrm->numfloordata, 4, 1, f);
    if (ptrm->numfloordata)
    {    
        fwrite(ptrm->floordata, ptrm->numfloordata, 2, f);
    }
    header.dir.s_floordata = ftell(f)-header.dir.o_floordata;
    header.dir.o_items = ftell(f);
    fwrite(&ptrm->numitems, 4, 1, f);
    if (ptrm->numitems)
    {    
        i = ptrm->itemlisthead;
        while (i)
        {
            fwrite(i->item, sizeof(item2_t), 1, f);
            i = i->next;
        }
    }
    header.dir.s_items = ftell(f)-header.dir.o_items;
    fseek(f, 0, SEEK_SET);
    fwrite(&header, sizeof(trl_header_t), 1, f);
    fclose(f);
    return TRUE;
}

#ifdef __cplusplus
}
#endif
