//tr2file.c

#ifdef __cplusplus
extern "C"
{
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <crtdbg.h>
#include "types.h"
#include "MyMemory.h"
#include "trdefs.h"
#include "tr2.h"
#include "patch.h"

#define ASSERT(e) _ASSERT(e)

#define THING 0x109

long LoadTr2File(char *lpszFileName, tr2_t *ptr2, unsigned __int64 loadflags)
{
#define ERR() { if (f != NULL) fclose(f); sprintf(szTemp, "failed on line %d\n", __LINE__); ASSERT(FALSE); return FALSE; }

#define RET() { if (f != NULL) fclose(f); return TRUE; }

    char szTemp[256];
    long size, pos;
    int x, c, num, dataptr;
    long l;
    short s, s2;
    FILE *f;

    f = fopen(lpszFileName, "rb");
    if (f == NULL)
    {    
        ERR();
    }
    
    fread(&ptr2->version, 4, 1, f);
    if (ptr2->version != TR2_VERSION)
    {    
        ERR();    
    }
    
    ptr2->offsets.palette = ftell(f);
    if (loadflags & TR2_PALETTE)
    {
        ptr2->palette = (u_char *)GetMemBlock(768);
        fread(ptr2->palette, 768, 1, f);
    }
    else
    {    
        fseek(f, 768, SEEK_CUR); //palette
    }

    //fixme?
    fseek(f, 0x400, SEEK_CUR);

    fread(&ptr2->numtextiles, 4, 1, f);
    if (ptr2->numtextiles <= 0)
    {    
        ERR();
    }
    ptr2->offsets.textiles8 = ftell(f);
    if (loadflags & TR2_TEXTURETILES8)
    {
        ptr2->textiles8 = (textile8_t *)GetMemBlock(ptr2->numtextiles*sizeof(textile8_t));
        fread(ptr2->textiles8, sizeof(textile8_t), ptr2->numtextiles, f);
    }
    else
    {    
        fseek(f, ptr2->numtextiles*sizeof(textile8_t), SEEK_CUR);
    }
    
    ptr2->offsets.textiles16 = ftell(f);
    if (loadflags & TR2_TEXTURETILES16)
    {
        ptr2->textiles16 = (textile16_t *)GetMemBlock(ptr2->numtextiles*sizeof(textile16_t));
        fread(ptr2->textiles16, sizeof(textile16_t), ptr2->numtextiles, f);
    }
    else
    {    
        fseek(f, ptr2->numtextiles*sizeof(textile16_t), SEEK_CUR);
    }

    fread(&l, 4, 1, f);
    if (l != 0)
    {    
        ERR();
    }
    
    fread(&ptr2->numrooms, 2, 1, f);
    if (ptr2->numrooms <= 0)
    {    
        ERR();
    }
    ptr2->offsets.rooms = ftell(f);
    //ptr2->roomoffsets = (long *)GetMemBlock(ptr2->numrooms*4);
    if (loadflags & TR2_ROOMS)
    {
        ptr2->rooms = (room2_t *)GetMemBlock(ptr2->numrooms*sizeof(room2_t));
        for (x = 0; x < ptr2->numrooms; x++)
        {
            //ptr2->roomoffsets[x] = ftell(f);
            pos = ftell(f);
            fread(&ptr2->rooms[x].roominfo, sizeof(roominfo_t), 1, f);
            pos = ftell(f);
    
            pos = ftell(f);
            fread(&ptr2->rooms[x].numrtwords, 4, 1, f);
            pos = ftell(f);

            ptr2->rooms[x].rtwords = (u_char *)GetMemBlock(ptr2->rooms[x].numrtwords*2);
            pos = ftell(f);
            fread(ptr2->rooms[x].rtwords, 2, ptr2->rooms[x].numrtwords, f);
            pos = ftell(f);
            /*dataptr = 0;
            
            ptr2->rooms[x].numvertices = *(short *)(ptr2->rooms[x].rtwords);
            dataptr += 2;
            if (ptr2->rooms[x].numvertices > 0)
                ptr2->rooms[x].vertices = (rmvertex_t *)(ptr2->rooms[x].rtwords+dataptr);
            dataptr += ptr2->rooms[x].numvertices*sizeof(rmvertex_t);
            
            ptr2->rooms[x].numrectangles = *(short *)(ptr2->rooms[x].rtwords+dataptr);
            dataptr += 2;
            if (ptr2->rooms[x].numrectangles > 0)
                ptr2->rooms[x].rectangles = (face4_t *)(ptr2->rooms[x].rtwords+dataptr);
            dataptr += ptr2->rooms[x].numrectangles*sizeof(face4_t);
            
            ptr2->rooms[x].numtriangles = *(short *)(ptr2->rooms[x].rtwords+dataptr);
            dataptr += 2;
            if (ptr2->rooms[x].numtriangles > 0)
                ptr2->rooms[x].triangles = (face3_t *)(ptr2->rooms[x].rtwords+dataptr);
            dataptr += ptr2->rooms[x].numtriangles*sizeof(face3_t);
            
            ptr2->rooms[x].numsprites = *(short *)(ptr2->rooms[x].rtwords+dataptr);
            dataptr += 2;
            ptr2->rooms[x].sprites = (roomsprite_t *)(ptr2->rooms[x].rtwords+dataptr);*/

            pos = ftell(f);
            fread(&ptr2->rooms[x].numdoors, 2, 1, f);
            pos = ftell(f);
            if (ptr2->rooms[x].numdoors > 0)
            {
                ptr2->rooms[x].doors = (roomdoor_t *)GetMemBlock(ptr2->rooms[x].numdoors*sizeof(roomdoor_t));
                pos = ftell(f);
                fread(ptr2->rooms[x].doors, sizeof(roomdoor_t), ptr2->rooms[x].numdoors, f);
                pos = ftell(f);
            }
            
            pos = ftell(f);
            fread(&ptr2->rooms[x].numysectors, 2, 1, f);
            pos = ftell(f);
            fread(&ptr2->rooms[x].numxsectors, 2, 1, f);
            num = ptr2->rooms[x].numysectors*ptr2->rooms[x].numxsectors;
            if (num <= 0)
            {    
                ERR();
            }
            ptr2->rooms[x].sectors = (sector_t *)GetMemBlock(num*sizeof(sector_t));
            pos = ftell(f);
            fread(ptr2->rooms[x].sectors, sizeof(sector_t), num, f);

            pos = ftell(f);
            fread(&ptr2->rooms[x].un1, 1, 1, f);
            pos = ftell(f);
            fread(&ptr2->rooms[x].lightlevel1, 1, 1, f);
            pos = ftell(f);
            fread(&ptr2->rooms[x].un4, 1, 1, f);
            pos = ftell(f);
            fread(&ptr2->rooms[x].lightlevel2, 1, 1, f);
            pos = ftell(f);
            
            fread(&s, 2, 1, f); //unknown, not a count

            pos = ftell(f);
            /*fread(&ptr2->rooms[x].numstatics, 2, 1, f);
            if (ptr2->rooms[x].numstatics > 0)
            {
                ptr2->rooms[x].statics = (roomstatic_t *)GetMemBlock(ptr2->rooms[x].numstatics*sizeof(roomstatic_t));
                pos = ftell(f);
                fread(ptr2->rooms[x].statics, sizeof(roomstatic_t), ptr2->rooms[x].numstatics, f);
            }*/
            fread(&ptr2->rooms[x].numlights, 2, 1, f);
            if (ptr2->rooms[x].numlights > 0)
            {
                ptr2->rooms[x].lights = (roomlight2_t *)GetMemBlock(ptr2->rooms[x].numlights*sizeof(roomlight2_t));
                pos = ftell(f);
                fread(ptr2->rooms[x].lights, sizeof(roomlight2_t), ptr2->rooms[x].numlights, f);
            }
            
            pos = ftell(f);
            fread(&s, 2, 1, f);
            fseek(f, s*20, SEEK_CUR); //what is this? statics?
            pos = ftell(f);
            
            pos = ftell(f);
            fread(&ptr2->rooms[x].un2, 2, 1, f);
            pos = ftell(f);
            fread(&ptr2->rooms[x].water, 1, 1, f);
            pos = ftell(f);
            fread(&ptr2->rooms[x].un3, 1, 1, f);
            pos = ftell(f);
            //fread(&l, 4, 1, f);
            //fseek(f, 4, SEEK_CUR);
            pos = ftell(f);
        }
    }
    else
    {
        for (x = 0; x < ptr2->numrooms; x++)
        {
            pos = ftell(f);
            //ptr2->roomoffsets[x] = ftell(f);
            fseek(f, sizeof(roominfo_t), SEEK_CUR);
            pos = ftell(f);
    
            pos = ftell(f);
            fread(&l, 4, 1, f);
            pos = ftell(f);
            pos = ftell(f);
            fseek(f, l*2, SEEK_CUR);

            /*fread(&s, 2, 1, f);
            fseek(f, s*12, SEEK_CUR);

            fread(&s, 2, 1, f);
            fseek(f, s*10, SEEK_CUR);

            fread(&s, 2, 1, f);
            fseek(f, s*8, SEEK_CUR);

            fread(&s, 2, 1, f);
            if (s != 0)
            {
                pos = ftell(f);
                sprintf(szTemp, "%d\n", pos);
                TRACE0(szTemp);
            }*/
            /*dataptr = 0;
            
            ptr2->rooms[x].numvertices = *(short *)(ptr2->rooms[x].rtwords);
            dataptr += 2;
            if (ptr2->rooms[x].numvertices > 0)
                ptr2->rooms[x].vertices = (rmvertex_t *)(ptr2->rooms[x].rtwords+dataptr);
            dataptr += ptr2->rooms[x].numvertices*sizeof(rmvertex_t);
            
            ptr2->rooms[x].numrectangles = *(short *)(ptr2->rooms[x].rtwords+dataptr);
            dataptr += 2;
            if (ptr2->rooms[x].numrectangles > 0)
                ptr2->rooms[x].rectangles = (face4_t *)(ptr2->rooms[x].rtwords+dataptr);
            dataptr += ptr2->rooms[x].numrectangles*sizeof(face4_t);
            
            ptr2->rooms[x].numtriangles = *(short *)(ptr2->rooms[x].rtwords+dataptr);
            dataptr += 2;
            if (ptr2->rooms[x].numtriangles > 0)
                ptr2->rooms[x].triangles = (face3_t *)(ptr2->rooms[x].rtwords+dataptr);
            dataptr += ptr2->rooms[x].numtriangles*sizeof(face3_t);
            
            ptr2->rooms[x].numsprites = *(short *)(ptr2->rooms[x].rtwords+dataptr);
            dataptr += 2;
            ptr2->rooms[x].sprites = (roomsprite_t *)(ptr2->rooms[x].rtwords+dataptr);*/

            pos = ftell(f);
            fread(&s, 2, 1, f);
            pos = ftell(f);
            if (s > 0)
            {
                pos = ftell(f);
                fseek(f, s*sizeof(roomdoor_t), SEEK_CUR);
                pos = ftell(f);
            }
            
            pos = ftell(f);
            fread(&s, 2, 1, f);
            pos = ftell(f);
            fread(&s2, 2, 1, f);
            num = s*s2;
            if (num <= 0)
            {    
                ERR();
            }
            pos = ftell(f);
            fseek(f, num*sizeof(sector_t), SEEK_CUR);

            pos = ftell(f);

            fseek(f, 4, SEEK_CUR);
            
            fread(&s, 2, 1, f); //unknown, not a count

            pos = ftell(f);
            /*fread(&ptr2->rooms[x].numstatics, 2, 1, f);
            if (ptr2->rooms[x].numstatics > 0)
            {
                ptr2->rooms[x].statics = (roomstatic_t *)GetMemBlock(ptr2->rooms[x].numstatics*sizeof(roomstatic_t));
                pos = ftell(f);
                fread(ptr2->rooms[x].statics, sizeof(roomstatic_t), ptr2->rooms[x].numstatics, f);
            }*/
            fread(&s, 2, 1, f);
            if (s > 0)
            {
                pos = ftell(f);
                fseek(f, s*sizeof(roomlight2_t), SEEK_CUR);
            }
            
            pos = ftell(f);
            fread(&s, 2, 1, f);
            fseek(f, s*20, SEEK_CUR); //what is this? statics?
            pos = ftell(f);
            
            pos = ftell(f);
            fseek(f, 4, SEEK_CUR);
            //fread(&l, 4, 1, f);
            //fseek(f, 4, SEEK_CUR);
            pos = ftell(f);
        }
    }
    
    ptr2->offsets.roomsize = ftell(f)-ptr2->offsets.rooms;

    fread(&ptr2->numfloordata, 4, 1, f);
    ptr2->offsets.floordata = ftell(f);
    if ((loadflags & TR2_FLOORDATA) && (ptr2->numfloordata > 0))
    {
        ptr2->floordata = (u_char *)GetMemBlock(ptr2->numfloordata*2);
        fread(ptr2->floordata, 2, ptr2->numfloordata, f);
    }
    else
    {    
        fseek(f, ptr2->numfloordata*2, SEEK_CUR);
    }
    
    pos = ftell(f);
    fread(&ptr2->nummeshwords, 4, 1, f);
    ptr2->offsets.meshes = ftell(f);
    if ((loadflags & TR2_MESHES) && (ptr2->nummeshwords > 0))
    {
        ptr2->meshwords = (short *)GetMemBlock(ptr2->nummeshwords*2);
        fread(ptr2->meshwords, 2, ptr2->nummeshwords, f);
    }
    else
    {    
        fseek(f, ptr2->nummeshwords*2, SEEK_CUR);
    }
    
    pos = ftell(f);
    fread(&ptr2->nummeshpointers, 4, 1, f);
    ptr2->offsets.meshpointers = ftell(f);
    if ((loadflags & TR2_MESHES) && (ptr2->nummeshpointers > 0))
    {
        ptr2->meshpointers = (long *)GetMemBlock(ptr2->nummeshpointers*4);
        fread(ptr2->meshpointers, 4, ptr2->nummeshpointers, f);
    }
    else
    {    
        fseek(f, ptr2->nummeshpointers*4, SEEK_CUR);
    }
    
    pos = ftell(f);
    fread(&ptr2->numanims, 4, 1, f); //anims
    if (ptr2->numanims == THING)
    {
        ASSERT(FALSE);
    }
    ptr2->offsets.anims = ftell(f);
    if ((loadflags & TR2_ANIMS) && (ptr2->numanims > 0))
    {
        ptr2->anims = (anim_t *)GetMemBlock(ptr2->numanims*sizeof(anim_t));
        fread(ptr2->anims, sizeof(anim_t), ptr2->numanims, f);
    }
    else
    {    
        fseek(f, ptr2->numanims*sizeof(anim_t), SEEK_CUR);
    }
    
    pos = ftell(f);
    fread(&ptr2->numstructs, 4, 1, f); //structs
    if (ptr2->numstructs == THING)
    {
        ASSERT(FALSE);
    }
    ptr2->offsets.structs = ftell(f);
    if ((loadflags & TR2_STRUCTS) && (ptr2->numstructs > 0))
    {
        ptr2->structs = (struct_t *)GetMemBlock(ptr2->numstructs*sizeof(struct_t));
        fread(ptr2->structs, sizeof(struct_t), ptr2->numstructs, f);
    }
    else
    {    
        fseek(f, ptr2->numstructs*sizeof(struct_t), SEEK_CUR);
    }
    
    pos = ftell(f);
    fread(&ptr2->numranges, 4, 1, f); //ranges
    if (ptr2->numranges == THING)
    {
        ASSERT(FALSE);
    }
    ptr2->offsets.ranges = ftell(f);
    if ((loadflags & TR2_RANGES) && (ptr2->numranges > 0))
    {
        ptr2->ranges = (range_t *)GetMemBlock(ptr2->numranges*sizeof(range_t));
        fread(ptr2->ranges, sizeof(range_t), ptr2->numranges, f);
    }
    else
    {    
        fseek(f, ptr2->numranges*sizeof(range_t), SEEK_CUR);
    }
    
    pos = ftell(f);
    fread(&ptr2->numbones1, 4, 1, f); //bones 1
    if (ptr2->numbones1 == THING)
    {
        ASSERT(FALSE);
    }
    ptr2->offsets.bones1 = ftell(f);
    if ((loadflags & TR2_BONES) && (ptr2->numbones1 > 0))
    {
        ptr2->bones1 = (bone1_t *)GetMemBlock(ptr2->numbones1*sizeof(bone1_t));
        fread(ptr2->bones1, sizeof(bone1_t), ptr2->numbones1, f);
    }
    else
    {    
        fseek(f, ptr2->numbones1*sizeof(bone1_t), SEEK_CUR);
    }
    
    pos = ftell(f);
    fread(&ptr2->numbones2, 4, 1, f); //bones 2
    if (ptr2->numbones2 == THING)
    {
        ASSERT(FALSE);
    }
    ptr2->offsets.bones2 = ftell(f);
    if ((loadflags & TR2_BONES) && (ptr2->numbones2 > 0))
    {
        ptr2->bones2 = (bone2_t *)GetMemBlock(ptr2->numbones2*sizeof(bone2_t));
        fread(ptr2->bones2, sizeof(bone2_t), ptr2->numbones2, f);
    }
    else
    {    
        fseek(f, ptr2->numbones2*sizeof(bone2_t), SEEK_CUR);
    }
    
    pos = ftell(f);
    fread(&ptr2->numframes, 4, 1, f); //frames
    if (ptr2->numframes == THING)
    {
        ASSERT(FALSE);
    }
    ptr2->offsets.frames = ftell(f);
    if ((loadflags & TR2_FRAMES) && (ptr2->numframes > 0))
    {
        ptr2->frames = (frame_t *)GetMemBlock(ptr2->numframes*sizeof(frame_t));
        fread(ptr2->frames, sizeof(frame_t), ptr2->numframes, f);
    }
    else
    {    
        fseek(f, ptr2->numframes*sizeof(frame_t), SEEK_CUR);
    }
    
    pos = ftell(f);
    fread(&ptr2->numun2, 4, 1, f); //un2
    if (ptr2->numun2 == THING)
    {
        ASSERT(FALSE);
    }
    ptr2->offsets.un2 = ftell(f);
    if ((loadflags & TR2_UN2) && (ptr2->numun2 > 0))
    {
        ptr2->un2 = (un2_t *)GetMemBlock(ptr2->numun2*sizeof(un2_t));
        fread(ptr2->un2, sizeof(un2_t), ptr2->numun2, f);
    }
    else
    {    
        fseek(f, ptr2->numun2*sizeof(un2_t), SEEK_CUR);
    }
    
    pos = ftell(f);
    fread(&ptr2->numun3, 4, 1, f);
    if (ptr2->numun3 == THING)
    {
        ASSERT(FALSE);
    }
    ptr2->offsets.un3 = ftell(f);
    if ((loadflags & TR2_UN3) && (ptr2->numun3 > 0))
    {
        ptr2->un3 = (un3_t *)GetMemBlock(ptr2->numun3*sizeof(un3_t));
        fread(ptr2->un3, sizeof(un3_t), ptr2->numun3, f);
    }
    else
    {    
        fseek(f, ptr2->numun3*32, SEEK_CUR);
    }
    
    pos = ftell(f);
    fread(&ptr2->numobjtextures, 4, 1, f); //object textures
    ptr2->offsets.objtextures = ftell(f);
    if ((loadflags & TR2_OBJTEXTURES) && (ptr2->numobjtextures > 0))
    {
        ptr2->objtextures = (objtexture_t *)GetMemBlock(ptr2->numobjtextures*sizeof(objtexture_t));
        fread(ptr2->objtextures, sizeof(objtexture_t), ptr2->numobjtextures, f);
    }
    else
    {    
        fseek(f, ptr2->numobjtextures*sizeof(objtexture_t), SEEK_CUR);
    }
    
    pos = ftell(f);
    fread(&ptr2->numspritetextures, 4, 1, f); //sprite textures
    ptr2->offsets.spritetextures = ftell(f);
    if ((loadflags & TR2_SPRITETEXTURES) && (ptr2->numspritetextures > 0))
    {
        ptr2->spritetextures = (spritetexture_t *)GetMemBlock(ptr2->numspritetextures*sizeof(spritetexture_t));
        fread(ptr2->spritetextures, sizeof(spritetexture_t), ptr2->numspritetextures, f);
    }
    else
    {    
        fseek(f, ptr2->numspritetextures*sizeof(spritetexture_t), SEEK_CUR);
    }
    
    pos = ftell(f);
    fread(&ptr2->numun4, 4, 1, f);
    if (ptr2->numun4 == THING)
    {
        ASSERT(FALSE);
    }
    ptr2->offsets.un4 = ftell(f);
    if ((loadflags & TR2_UN4) && (ptr2->numun4 > 0))
    {
        ptr2->un4 = (un4_t *)GetMemBlock(ptr2->numun4*sizeof(un4_t));
        fread(ptr2->un4, sizeof(un4_t), ptr2->numun4, f);
    }
    else
    {    
        fseek(f, ptr2->numun4*sizeof(un4_t), SEEK_CUR);
    }
    
    pos = ftell(f);
    //fixme
    /*if (nFileType == FILETYPE_TR2_DEMO)
    {
        ptr2->offsets.colormaps = ftell(f);
        if (loadflags & TR2_COLORMAPS)
        {
            ptr2->colormaps = (colormap_t *)GetMemBlock(sizeof(colormap_t)*TR2_NUMCOLORMAPS);
            fread(ptr2->colormaps, sizeof(colormap_t), TR2_NUMCOLORMAPS, f);
        }
        else
            fseek(f, TR2_NUMCOLORMAPS*sizeof(colormap_t), SEEK_CUR);
    }*/

    pos = ftell(f);
    fread(&ptr2->numcameras, 4, 1, f); //cameras
    if (ptr2->numcameras == THING)
    {
        ASSERT(FALSE);
    }
    ptr2->offsets.cameras = ftell(f);
    if ((loadflags & TR2_CAMERAS) && (ptr2->numcameras > 0))
    {
        ptr2->cameras = (camera_t *)GetMemBlock(ptr2->numcameras*sizeof(camera_t));
        pos = ftell(f);
        fread(ptr2->cameras, sizeof(camera_t), ptr2->numcameras, f);
    }
    else
    {    
        fseek(f, ptr2->numcameras*sizeof(camera_t), SEEK_CUR);
    }
    
    pos = ftell(f);
    fread(&ptr2->numsoundfx, 4, 1, f); //sound fx
    if (ptr2->numsoundfx == THING)
    {
        ASSERT(FALSE);
    }
    ptr2->offsets.soundfx = ftell(f);
    if ((loadflags & TR2_SOUNDFX) && (ptr2->numsoundfx > 0))
    {
        ptr2->soundfx = (soundfx_t *)GetMemBlock(ptr2->numsoundfx*sizeof(soundfx_t));
    pos = ftell(f);
        fread(ptr2->soundfx, sizeof(soundfx_t), ptr2->numsoundfx, f);
    }
    else
    {    
        fseek(f, ptr2->numsoundfx*sizeof(soundfx_t), SEEK_CUR);
    }
    
    //long 8, boxes?
    fread(&ptr2->numboxes, 4, 1, f);
    if (ptr2->numboxes == THING)
    {
        ASSERT(FALSE);
    }
    ptr2->offsets.boxes1 = ftell(f);
    fseek(f, ptr2->numboxes*8, SEEK_CUR);

    pos = ftell(f);
    fread(&ptr2->numoverlaps, 4, 1, f); //overlaps
    if (ptr2->numoverlaps == THING)
    {
        ASSERT(FALSE);
    }
    ptr2->offsets.overlaps = ftell(f);
    if ((loadflags & TR2_OVERLAPS) && (ptr2->numoverlaps > 0))
    {
        ptr2->overlaps = (overlap_t *)GetMemBlock(ptr2->numoverlaps*sizeof(overlap_t));
    pos = ftell(f);
        fread(ptr2->overlaps, sizeof(overlap_t), ptr2->numoverlaps, f);
    }
    else
    {    
        fseek(f, ptr2->numoverlaps*sizeof(overlap_t), SEEK_CUR);
    }
        
    ptr2->offsets.boxes2 = ftell(f);
    fseek(f, ptr2->numboxes*20, SEEK_CUR);

    pos = ftell(f);
    fread(&ptr2->numanimtextures, 4, 1, f); //animating texture ranges
    ptr2->offsets.animtextures = ftell(f);
    if ((loadflags & TR2_ANIMTEXTURES) && (ptr2->numanimtextures > 0))
    {
        ptr2->animtextures = (animtexture_t *)GetMemBlock(ptr2->numanimtextures*sizeof(animtexture_t));
    pos = ftell(f);
        fread(ptr2->animtextures, sizeof(animtexture_t), ptr2->numanimtextures, f);
    }
    else
    {    
        fseek(f, ptr2->numanimtextures*sizeof(animtexture_t), SEEK_CUR);
    }
    
    pos = ftell(f);
    fread(&ptr2->numitems, 4, 1, f); //items
    ptr2->offsets.items = ftell(f);
    if ((loadflags & TR2_ITEMS) && (ptr2->numitems > 0))
    {
        ptr2->items = (item2_t *)GetMemBlock(sizeof(item2_t)*ptr2->numitems);
    pos = ftell(f);
        fread(ptr2->items, sizeof(item2_t), ptr2->numitems, f);
    }
    else
    {    
        fseek(f, ptr2->numitems*sizeof(item2_t), SEEK_CUR);
    }
    
    pos = ftell(f);
    //if (nFileType == FILETYPE_TR2_RETAIL)
    {
        ptr2->offsets.colormaps = ftell(f);
        if (loadflags & TR2_COLORMAPS)
        {
            ptr2->colormaps = (colormap_t *)GetMemBlock(sizeof(colormap_t)*TR2_NUMCOLORMAPS);
            fread(ptr2->colormaps, sizeof(colormap_t), TR2_NUMCOLORMAPS, f);
        }
        else
        {    
            fseek(f, TR2_NUMCOLORMAPS*sizeof(colormap_t), SEEK_CUR);
        }
    }

    pos = ftell(f);
    fread(&ptr2->numun6, 2, 1, f); //?
    if (ptr2->numun6 == THING)
    {
        ASSERT(FALSE);
    }
    ptr2->offsets.un6 = ftell(f);
    if ((loadflags & TR2_UN6) && (ptr2->numun6 > 0))
    {
        ptr2->un6 = (un6_t *)GetMemBlock(ptr2->numun6*sizeof(un6_t));
    pos = ftell(f);
        fread(ptr2->un6, sizeof(un6_t), ptr2->numun6, f);
    }
    else
    {    
        fseek(f, ptr2->numun6*sizeof(un6_t), SEEK_CUR);
    }
    
    pos = ftell(f);
    fread(&ptr2->demodatasize, 2, 1, f); //demo data
    ptr2->offsets.demodata = ftell(f);
    if ((loadflags & TR2_DEMODATA) && (ptr2->demodatasize > 0))
    {
        ptr2->demodata = (u_char *)GetMemBlock(ptr2->demodatasize);
    pos = ftell(f);
        fread(ptr2->demodata, ptr2->demodatasize, 1, f);
    }
    else
    {    
        fseek(f, ptr2->demodatasize, SEEK_CUR);
    }
    
    ptr2->offsets.soundstuff = ftell(f);
    if (loadflags & TR2_SOUNDSTUFF)
    {
        ptr2->soundstuff = (u_char *)GetMemBlock(740);
    pos = ftell(f);
        fread(ptr2->soundstuff, 740, 1, f);
    }
    else
    {    
        fseek(f, 740, SEEK_CUR); //sound stuff
    }
    
    pos = ftell(f);
    fread(&ptr2->numsampleinfos, 4, 1, f); //sample infos
    ptr2->offsets.sampleinfos = ftell(f);
    if ((loadflags & TR2_SAMPLEINFOS) && (ptr2->numsampleinfos > 0))
    {
        ptr2->sampleinfos = (sampleinfo_t *)GetMemBlock(sizeof(sampleinfo_t)*ptr2->numsampleinfos);
    pos = ftell(f);
        fread(ptr2->sampleinfos, sizeof(sampleinfo_t), ptr2->numsampleinfos, f);
    }
    else
    {    
        fseek(f, ptr2->numsampleinfos*sizeof(sampleinfo_t), SEEK_CUR);
    }
    
    pos = ftell(f);
    fread(&ptr2->numsamples, 4, 1, f);
    ptr2->offsets.samples = ftell(f);
    //fixme
    //if (nFileType == FILETYPE_TR2_RETAIL)
    {
        //ptr2->sampleindexes = (long *)GetMemBlock(ptr2->numsamples*4);
        ptr2->sampleindexes = (long *)malloc(ptr2->numsamples*4);
        fread(ptr2->sampleindexes, 4, ptr2->numsamples, f);
    }
    /*pos = ftell(f);
    fread(&l, 4, 1, f); //samples
    if ((loadflags & TR2_SAMPLES) && (l > 0))
    {
        ptr2->samplesize = l;
        ptr2->samples = (u_char *)GetMemBlock(ptr2->samplesize);
    pos = ftell(f);
        fread(ptr2->samples, ptr2->samplesize, 1, f);
    }
    else
        fseek(f, l, SEEK_CUR);
    
    pos = ftell(f);
    fread(&l, 4, 1, f); //sample offsets
    if ((loadflags & TR2_SAMPLEOFFSETS) && (l > 0))
    {
        ptr2->numsampleoffsets = l;
        ptr2->sampleoffsets = (long *)GetMemBlock(4*ptr2->numsampleoffsets);
    pos = ftell(f);
        fread(ptr2->sampleoffsets, 4, ptr2->numsampleoffsets, f);
    }
    else
        fseek(f, l*4, SEEK_CUR);
    

    fseek(f, 0, SEEK_END);
    size = ftell(f);
    ASSERT(pos == size);*/
    
    fclose(f);
    return TRUE;
}

long BuildTr2File(char *lpszDestTr2, char *lpszSrcTrl, char *lpszBasePath, tr2_t *ptr2)
{
#define ERR2() { free(buffer); if (trl != NULL) fclose(trl); if (f != NULL) fclose(f); if (base != NULL) fclose(base); sprintf(szTemp, "failed on line %d\n", __LINE__); ASSERT(FALSE); return FALSE; }

    char szPath[256], szTemp[256];
    trl_header_t header;
    u_char *buffer;
    FILE *f = NULL, *base = NULL, *trl = NULL, *f2 = NULL;
    long *sfxofs;
    u_char *dataptr;
    int x;
    long l, basesize, sfxofssize, sfxsize, sampleindex;
    short s;

    f = fopen(lpszDestTr2, "wb");
    if (f == NULL)
    {    
        ERR2();
    }
    
    trl = fopen(lpszSrcTrl, "rb");
    fread(&header, sizeof(trl_header_t), 1, trl);

    sprintf(szPath, "%s\\%s", lpszBasePath, header.baselevel);
    base = fopen(szPath, "rb");
    if (base == NULL)
    {    
        ERR2();
    }
    
    fseek(base, 0, SEEK_END);
    basesize = ftell(base);
    fseek(base, 0, SEEK_SET);
    
    buffer = (u_char *)malloc(PATCHFILE_BUFSIZE);
    
    l = TR2_VERSION;
    fwrite(&l, 4, 1, f);
    
    //palette
    PatchFile(buffer, f, base, 4, 768);
    
    //unknown
    PatchFile(buffer, f, base, 4+768, 0x400);
    
    fwrite(&ptr2->numtextiles, 4, 1, f);
    PatchFile(buffer, f, base, ptr2->offsets.textiles8, ptr2->numtextiles*sizeof(textile8_t));
    PatchFile(buffer, f, base, ptr2->offsets.textiles16, ptr2->numtextiles*sizeof(textile16_t));
    
    //unknown, always 0
    l = 0;
    fwrite(&l, 4, 1, f);
    
    /*s = 1;
    fwrite(&s, 2, 1, f);
    proom->roominfo.x = 0x7400;
    proom->roominfo.z = 0x7400;
    fwrite(&proom->roominfo, sizeof(roominfo_t), 1, f);

    fwrite(&proom->nummeshwords, 4, 1, f);
    ASSERT(proom->nummeshwords);
    fwrite(&proom->mesh.numvertices, 2, 1, f);
    if (proom->mesh.numvertices > 0)
        fwrite(proom->mesh.vertices, sizeof(rmvertex2_t), proom->mesh.numvertices, f);
    fwrite(&proom->mesh.numpoly4s, 2, 1, f);
    if (proom->mesh.poly4s > 0)
        fwrite(proom->mesh.poly4s, sizeof(poly4_t), proom->mesh.numpoly4s, f);
    fwrite(&proom->mesh.numpoly3s, 2, 1, f);
    if (proom->mesh.poly3s > 0)
        fwrite(proom->mesh.poly3s, sizeof(poly3_t), proom->mesh.numpoly3s, f);
    fwrite(&proom->mesh.un1, 2, 1, f);*/

        //fwrite(proom->meshwords, 2, proom->nummeshwords, f);

    //proom->numrtwords = 0x9f6;
    //PatchFile(buffer, f, base, ptr2->offsets.rooms+sizeof(roominfo_t)+4, proom->numrtwords*2);
    

    /*fwrite(&proom->numdoors, 2, 1, f);
    if (proom->numdoors > 0)
        fwrite(proom->doors, sizeof(roomdoor_t), proom->numdoors, f);
    fwrite(&proom->numysectors, 2, 1, f);
    fwrite(&proom->numxsectors, 2, 1, f);
    //fwrite(proom->sectors, sizeof(sector_t), proom->numxsectors*proom->numysectors, f);
    for (x = 0; x < proom->numxsectors*proom->numysectors; x++)
        fwrite(&proom->sectors[x], sizeof(sector_t), 1, f);
    fwrite(&proom->un1, 1, 1, f);
    fwrite(&proom->lightlevel1, 1, 1, f);
    fwrite(&proom->un4, 1, 1, f);
    fwrite(&proom->lightlevel2, 1, 1, f);
    s = 0;
    fwrite(&s, 2, 1, f);
    fwrite(&s, 2, 1, f);
    fwrite(&s, 2, 1, f);

    proom->un2 = 0xffff;
    fwrite(&proom->un2, 2, 1, f);
    fwrite(&proom->water, 1, 1, f);
    fwrite(&proom->un3, 1, 1, f);*/

    PatchFile(buffer, f, trl, header.dir.o_rooms, header.dir.s_rooms);

    //floordata
    //fwrite(&ptr2->numfloordata, 4, 1, f);
    //PatchFile(buffer, f, base, ptr2->offsets.floordata, ptr2->numfloordata*2);
    //l = 0;
    //fwrite(&l, 4, 1, f);
    PatchFile(buffer, f, trl, header.dir.o_floordata, header.dir.s_floordata);


    fwrite(&ptr2->nummeshwords, 4, 1, f);
    PatchFile(buffer, f, base, ptr2->offsets.meshes, ptr2->nummeshwords*2);
    
    fwrite(&ptr2->nummeshpointers, 4, 1, f);
    PatchFile(buffer, f, base, ptr2->offsets.meshpointers, ptr2->nummeshpointers*4);
    
    fwrite(&ptr2->numanims, 4, 1, f);
    PatchFile(buffer, f, base, ptr2->offsets.anims, ptr2->numanims*sizeof(anim_t));
    
    fwrite(&ptr2->numstructs, 4, 1, f);
    PatchFile(buffer, f, base, ptr2->offsets.structs, ptr2->numstructs*sizeof(struct_t));
    
    fwrite(&ptr2->numranges, 4, 1, f);
    PatchFile(buffer, f, base, ptr2->offsets.ranges, ptr2->numranges*sizeof(range_t));
    
    fwrite(&ptr2->numbones1, 4, 1, f);
    PatchFile(buffer, f, base, ptr2->offsets.bones1, ptr2->numbones1*sizeof(bone1_t));
    
    fwrite(&ptr2->numbones2, 4, 1, f);
    PatchFile(buffer, f, base, ptr2->offsets.bones2, ptr2->numbones2*sizeof(bone2_t));
    
    fwrite(&ptr2->numframes, 4, 1, f);
    PatchFile(buffer, f, base, ptr2->offsets.frames, ptr2->numframes*sizeof(frame_t));
    
    //0 for un2 will not show lara at all (or doesn't appear to) other meshes are just
    //shown as first lara mesh, also doesn't show sky, and some things (switches,
    //doors?) don't work
    fwrite(&ptr2->numun2, 4, 1, f);
    PatchFile(buffer, f, base, ptr2->offsets.un2, ptr2->numun2*sizeof(un2_t));
    
    //un3 does nothing?
    fwrite(&ptr2->numun3, 4, 1, f);
    PatchFile(buffer, f, base, ptr2->offsets.un3, ptr2->numun3*sizeof(un3_t));
    //l = 0;
    //fwrite(&l, 4, 1, f);
    
    fwrite(&ptr2->numobjtextures, 4, 1, f);
    PatchFile(buffer, f, base, ptr2->offsets.objtextures, ptr2->numobjtextures*sizeof(objtexture_t));
    
    fwrite(&ptr2->numspritetextures, 4, 1, f);
    PatchFile(buffer, f, base, ptr2->offsets.spritetextures, ptr2->numspritetextures*sizeof(spritetexture_t));
    
    //un4 does nothing?
    fwrite(&ptr2->numun4, 4, 1, f);
    PatchFile(buffer, f, base, ptr2->offsets.un4, ptr2->numun4*sizeof(un4_t));
    //l = 0;
    //fwrite(&l, 4, 1, f);
    
    //if (nDestFileType == FILETYPE_TR2_DEMO)
       PatchFile(buffer, f, base, ptr2->offsets.colormaps, TR2_NUMCOLORMAPS*sizeof(colormap_t));
    
    fwrite(&ptr2->numcameras, 4, 1, f);
    PatchFile(buffer, f, base, ptr2->offsets.cameras, ptr2->numcameras*sizeof(camera_t));
    //l = 1;
    //fwrite(&l, 4, 1, f);
    //PatchFile(buffer, f, base, ptr2->offsets.cameras, l*sizeof(camera_t));

    fwrite(&ptr2->numsoundfx, 4, 1, f);
    PatchFile(buffer, f, base, ptr2->offsets.soundfx, ptr2->numsoundfx*sizeof(soundfx_t));
    //l = 0;
    //fwrite(&l, 4, 1, f);

    //no boxes makes camera not work as good, also screws up monster ai
    //fwrite(&ptr2->numboxes, 4, 1, f);
    //PatchFile(buffer, f, base, ptr2->offsets.boxes1, ptr2->numboxes*8);
    l = 0;
    fwrite(&l, 4, 1, f);
    
    //fwrite(&ptr2->numoverlaps, 4, 1, f);
    //PatchFile(buffer, f, base, ptr2->offsets.overlaps, ptr2->numoverlaps*sizeof(overlap_t));
    l = 0;
    fwrite(&l, 4, 1, f);
    
    //PatchFile(buffer, f, base, ptr2->offsets.boxes2, ptr2->numboxes*20);
    
    fwrite(&ptr2->numanimtextures, 4, 1, f);
    PatchFile(buffer, f, base, ptr2->offsets.animtextures, ptr2->numanimtextures*sizeof(animtexture_t));
    
    //fwrite(&ptr2->numitems, 4, 1, f);
    //PatchFile(buffer, f, base, ptr2->offsets.items, ptr2->numitems*sizeof(item2_t));
    /*l = 1;
    long blah = ftell(f);
    fwrite(&l, 4, 1, f);
    item2_t *items;
    items = (item2_t *)malloc(ptr2->numitems*sizeof(item2_t));
    fseek(base, ptr2->offsets.items, SEEK_SET);
    fread(items, sizeof(item2_t), ptr2->numitems, base);
    //for (x = 0; x < ptr2->numitems; x++)
        //items[x].light = 0;
    items[0].light = 0;
    fwrite(&items[0], sizeof(item2_t), 1, f);
    free(items);*/

    PatchFile(buffer, f, trl, header.dir.o_items, header.dir.s_items);

    //if (nDestFileType == FILETYPE_TR2_RETAIL)
        //PatchFile(buffer, f, base, ptr2->offsets.colormaps, TR2_NUMCOLORMAPS*sizeof(colormap_t));
    
    fwrite(&ptr2->numun6, 2, 1, f);
    PatchFile(buffer, f, base, ptr2->offsets.un6, ptr2->numun6*sizeof(un6_t));
    
    //fwrite(&ptr2->demodatasize, 2, 1, f);
    //PatchFile(buffer, f, base, ptr2->offsets.demodata, ptr2->demodatasize);
    s = 0;
    fwrite(&s, 2, 1, f);
    
    PatchFile(buffer, f, base, ptr2->offsets.soundstuff, 740);
    
    fwrite(&ptr2->numsampleinfos, 4, 1, f);
    PatchFile(buffer, f, base, ptr2->offsets.sampleinfos, ptr2->numsampleinfos*sizeof(sampleinfo_t));
    
    //samples
    fwrite(&ptr2->numsamples, 4, 1, f);
    //if ((nDestFileType == FILETYPE_TR2_DEMO) && (nSrcFileType == FILETYPE_TR2_RETAIL))
    {
        //fixme: use correct sfxofs
        f2 = fopen("c:\\sfxofs.dat", "rb");
        fseek(f2, 0, SEEK_END);
        sfxofssize = ftell(f2);
        fseek(f2, 0, SEEK_SET);
        sfxofs = (long *)malloc(sfxofssize);
        fread(sfxofs, sfxofssize, 1, f2);
        fclose(f2);
        strcpy(szPath, lpszBasePath);
        if (szPath[strlen(szPath)-1] != '\\')
        {
            strcat(szPath, "\\");
        }
        strcat(szPath, "main.sfx");
        f2 = fopen(szPath, "rb");
        fseek(f2, 0, SEEK_END);
        sfxsize = ftell(f2);
        fseek(f2, 0, SEEK_SET);
        for (x = 0; x < ptr2->numsamples; x++)
        {
            sampleindex = ptr2->sampleindexes[x];
            if (sampleindex < (sfxofssize/4)-1)
            {    
                PatchFile(buffer, f, f2, sfxofs[sampleindex], sfxofs[sampleindex+1]-sfxofs[sampleindex]);
            }
            else
            {    
                PatchFile(buffer, f, f2, sfxofs[sampleindex], sfxsize-sfxofs[sampleindex]);
            }
        }
        fclose(f2);
    }
    //else
        //PatchFile(buffer, f, base, ptr2->offsets.samples, srcsize-ptr2->offsets.samples);
    
    free(buffer);
    fclose(base);
    fclose(f);
    return TRUE;
}

#ifdef __cplusplus
}
#endif
